/*
 * usbrhctl.c -- control Strawberry Linux USBRH
 */

/*-
 * Copyright (C) 2008-2015 Nobuyuki Honda
 *
 *  This software is provided 'as-is', without any express or implied
 *  warranty.  In no event will the authors be held liable for any damages
 *  arising from the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, subject to the following restrictions:
 *
 *  1. The origin of this software must not be misrepresented; you must not
 *     claim that you wrote the original software. If you use this software
 *     in a product, an acknowledgment in the product documentation would be
 *     appreciated but is not required.
 *  2. Altered source versions must be plainly marked as such, and must not be
 *     misrepresented as being the original software.
 *  3. This notice may not be removed or altered from any source distribution.
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/sysctl.h>
#include <dev/usb/usb.h>
#include <dev/usb/usb_ioctl.h>
#include <dev/usb/usbhid.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#define VENDOR 0x1774
#define PRODUCT 0x1001

static int	findusbrh(char *dev, size_t size);
static void	calcSHT11(float *humi, float *temp);
static int	readsensor(int d, float *humi, float *temp);
static int	switchled(int d, int led, int on);
static int	switchheater(int d, int on);
static void	usage(void);

static int
findusbrh(char *dev, size_t size)
{
	int i;
	char uhid[PATH_MAX];
	char tmp[256];
	char *buf;
	size_t len;

	for (i = 0; i < USB_MAX_DEVICES; i++) {
		snprintf(uhid, sizeof(uhid), "/dev/uhid%d", i);
		if (strlen(dev) && strcmp(uhid, dev) != 0)
			continue;
		snprintf(tmp, sizeof(tmp), "dev.uhid.%d.%%pnpinfo", i);
		if (sysctlbyname(tmp, NULL, &len, NULL, 0) != 0) {
			if (strlen(dev)) {
				warn("sysctlbyname: '%s'", tmp);
				break;
			}
			continue;
		}
		buf = malloc(len);
		if (!buf)
			err(1, NULL);
		if (sysctlbyname(tmp, buf, &len, NULL, 0) != 0) {
			free(buf);
			warn("sysctlbyname: '%s'", tmp);
			if (strlen(dev))
				break;
			continue;
		}
		snprintf(tmp, sizeof(tmp), "vendor=%#x product=%#x ",
			VENDOR, PRODUCT);
		if (strstr(buf, tmp) == NULL) {
			free(buf);
			if (strlen(dev))
				break;
			continue;
		}
		free(buf);
		if (!strlen(dev))
			strlcpy(dev, uhid, size);
		return (1);
	}
	return (0);
}

static void
calcSHT11(float *humi, float *temp)
{
	const float	C1 = -4.0;
	const float	C2 = 0.0405;
	const float	C3 = -0.0000028;
	const float	T1 = 0.01;
	const float	T2 = 0.00008;
	const float	D1 = -40.0;
	const float	D2 = 0.01;
	float		t, rh, rh_lin, rh_true;

	t = *temp * D2 + D1;
	rh = *humi;
	rh_lin = C3 * rh * rh + C2 * rh + C1;
	rh_true = (t - 25) * (T1 + T2 * rh) + rh_lin;
	*humi = rh_true;
	*temp = t;
}

static int
readsensor(int d, float *humi, float *temp)
{
	struct usb_gen_descriptor ugd;
	u_char buf[8];
        char data[1024];

	memset(&ugd, 0, sizeof(ugd));
        ugd.ugd_data = data;
        ugd.ugd_maxlen = sizeof(data);
	ugd.ugd_report_type = UHID_OUTPUT_REPORT;
	if (ioctl(d, USB_SET_REPORT, &ugd) == -1) {
		warn("USB_SET_REPORT");
		return (1);
	}
	usleep(500*1000);
	ugd.ugd_report_type = UHID_OUTPUT_REPORT;
	if (ioctl(d, USB_SET_REPORT, &ugd) == -1) {
		warn("USB_SET_REPORT");
		return (1);
	}
	usleep(500*1000);

	if (read(d, buf, sizeof(buf)) == -1) {
		warn("read");
		return (1);
	}
	*humi = buf[0] << 8 | buf[1];
	*temp = buf[2] << 8 | buf[3];
	calcSHT11(humi, temp);
	return (0);
}

static int
switchled(int d, int led, int on)
{
	struct usb_gen_descriptor ugd;
        char data[2];

        data[0] = 3 + led;
        data[1] = on;
	memset(&ugd, 0, sizeof(ugd));
        ugd.ugd_data = data;
        ugd.ugd_maxlen = sizeof(data);
	ugd.ugd_report_type = UHID_FEATURE_REPORT;
	if (ioctl(d, USB_SET_REPORT, &ugd) == -1) {
		warn("USB_SET_REPORT");
		return (1);
	}
	return (0);
}

static int
switchheater(int d, int on)
{
	struct usb_gen_descriptor ugd;
        char data[2];

	data[0] = 1;
	data[1] = on << 2;
	memset(&ugd, 0, sizeof(ugd));
        ugd.ugd_data = data;
        ugd.ugd_maxlen = sizeof(data);
	ugd.ugd_report_type = UHID_FEATURE_REPORT;
	if (ioctl(d, USB_SET_REPORT, &ugd) == -1) {
		warn("USB_SET_REPORT");
		return (1);
	}
	return (0);
}

static void
usage(void)
{
	fprintf(stderr, "usage: usbrhctl [-din] [-f device]"
	    " [led0 | led1 | heater [on | off]]\n");
	exit(EX_USAGE);
}

int
main(int argc, char *argv[])
{
	int ch, d, dflag, iflag, nflag, on, ret;
	float humi, temp;
	const char *cmd;
	char dev[PATH_MAX];

	dflag = 0;
	iflag = 0;
	nflag = 0;
	on = 0;
	cmd = NULL;
	memset(dev, 0, sizeof(0));
	while ((ch = getopt(argc, argv, "df:in")) != -1) {
		switch (ch) {
		case 'd':
			dflag = 1;
			break;
		case 'f':
			strlcpy(dev, optarg, sizeof(dev));
			break;
		case 'i':
			iflag = 1;
			break;
		case 'n':
			nflag = 1;
			break;
		default:
			usage();
			/* NOTREACHED */
		}
	}
	argc -= optind;
	argv += optind;

	if (argc > 0) {
		cmd = argv[0];
		if (argc > 1)
			on = strcmp(argv[1], "on") == 0;
		else
			on = 1;
	}

	if (nflag && !strlen(dev))
		errx(1, "Not specified device");

	if (!nflag && !findusbrh(dev, sizeof(dev))) {
		if (!strlen(dev))
			errx(1, "USBRH not found");
		else
			errx(1, "%s: Not USBRH", dev);
	}

	d = open(dev, O_RDWR);
	if (d == -1)
		err(1, "%s", dev);

	if (cmd) {
		if (strcmp(cmd, "led0") == 0)
			ret = switchled(d, 0, on);
		else if (strcmp(cmd, "led1") == 0)
			ret = switchled(d, 1, on);
		else if (strcmp(cmd, "heater") == 0)
			ret = switchheater(d, on);
		else {
			warnx("Unknown command '%s'", cmd);
			ret = 1;
		}
	} else {
		ret = readsensor(d, &humi, &temp);
		if (ret == 0) {
			if (dflag)
				printf("%s: ", dev);
			if (iflag)
				printf("%d %d\n", (int)(temp + 0.5),
				    (int)(humi + 0.5));
			else
				printf("%.2f %.1f\n", temp, humi);
		}
	}
	return (ret);
}
